<?php

namespace Drupal\stacked_permissions\Form;

use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\user\PermissionHandler;

class AllowedPermissionsForm extends ConfigFormBase
{

    /**
     * {@inheritdoc}
     */
    protected function getEditableConfigNames()
    {
        return ['stacked_permissions.allowed_permissions'];
    }

    /**
     * {@inheritdoc}
     */
    public function getFormId()
    {
        return 'allowed_permissions_form';
    }

    /**
     * {@inheritdoc}
     */
    public function buildForm(array $form, FormStateInterface $form_state)
    {
        $config = $this->config('stacked_permissions.allowed_permissions');
        $allowed_permissions = $config->get('permissions') ?: [];
        $permission_handler = \Drupal::service('user.permissions');
        $permissions = $permission_handler->getPermissions();

        $grouped_permissions = [];
        foreach ($permissions as $permission_id => $permission) {
            $module = $permission['provider'];
            $module_name = \Drupal::service('extension.list.module')->getName($module);
            // Use -_- to replace spaces.
            $safe_permission_id = str_replace(' ', '-_-', $permission_id);
            $grouped_permissions[$module_name . '_|_' . $module][$safe_permission_id] = $permission;
        }

        foreach ($grouped_permissions as $module_name_concat => $permissions) {
            $module_name = explode('_|_', $module_name_concat)[0];
            $module = explode('_|_', $module_name_concat)[1];
            $form[$module_name] = [
                '#type' => 'details',
                '#title' => $this->t($module_name),
                '#open' => FALSE,
            ];

            foreach ($permissions as $permission_id => $permission) {
                $form[$module_name][$module . '_|_' . $permission_id] = [
                    '#type' => 'checkbox',
                    '#title' => $permission['title'],
                    '#default_value' => array_key_exists($module . '_|_' . $permission_id, $allowed_permissions),
                ];
            }
        }

        return parent::buildForm($form, $form_state);
    }

    /**
     * {@inheritdoc}
     */
    public function submitForm(array &$form, FormStateInterface $form_state)
    {
        $selected_permissions = [];

        // Process form values.
        foreach ($form_state->getValues() as $id => $checked) {
            if (str_contains($id, '_|_')) {
                $module = explode('_|_', $id)[0];
                $module_name = \Drupal::service('extension.list.module')->getName($module);

                if (!empty($module_name) && $checked) {
                    $selected_permissions[$id] = $checked;
                }
            }
        }

        // Save the selected permissions into the configuration.
        $this->config('stacked_permissions.allowed_permissions')
            ->set('permissions', $selected_permissions)
            ->save();

        // Remove disallowed permissions from all PermissionGroup entities.
        $entity_type_manager = \Drupal::entityTypeManager();
        $permission_group_storage = $entity_type_manager->getStorage('permission_group');
        $permission_groups = $permission_group_storage->loadMultiple();

        foreach ($permission_groups as $group) {
            $original_permissions = $group->get('permissions')->value;
            $unserialized = unserialize($original_permissions) ?: [];

            // Filter to keep only currently allowed permissions.
            $filtered = array_filter(
                $unserialized,
                function ($value, $perm_id) use ($selected_permissions) {
                    return isset($selected_permissions[$perm_id]);
                },
                ARRAY_FILTER_USE_BOTH
            );

            // If any were removed, save the updated permissions.
            if ($filtered !== $unserialized) {
                $group->set('permissions', serialize($filtered));
                $group->save();
                //Invalidate each group specific cache.
                $tags = [];
                if(is_array($group->getCacheTags())) {
                    $tags = array_merge($tags, $group->getCacheTags());
                } else {
                    $tags[] = $group->getCacheTags();
                }
                \Drupal::service('cache_tags.invalidator')->invalidateTags($tags);
            }
        }

        //Invalidate this config form.
        \Drupal::service('cache_tags.invalidator')->invalidateTags(['config:stacked_permissions.allowed_permissions']);

        parent::submitForm($form, $form_state);
    }

}