<?php

namespace Drupal\stacked_permissions\Form;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Url;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\user\Entity\User;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\InvokeCommand;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Ajax\HtmlCommand;
use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Render\Element;
use Drupal\Core\Render\Markup;
use Drupal\user\PermissionHandler;
use Drupal\Core\Extension\ModuleHandlerInterface;

class PermissionGroupForm extends ContentEntityForm {

    /**
     * {@inheritdoc}
     */
    public function buildForm(array $form, FormStateInterface $form_state) {
        // Clear the render cache to ensure UI reflects updated configuration.
        \Drupal::service('cache.render')->invalidateAll();

        $form = parent::buildForm($form, $form_state);

        // Insert a horizontal line followed by a header
        $form['add_group_membership_separator'] = [
            '#markup' => '<hr><h2>' . $this->t('Add a Group Member:') . '</h2>',
        ];

        // Add a container for messages.
        $form['custom_messages_add'] = [
            '#type' => 'container',
            '#attributes' => ['id' => 'custom-messages-add'],
        ];

        // Search for members to add
        $form['search_member'] = [
            '#type' => 'textfield',
            '#title' => $this->t('Search for User to Add'),
            '#description' => $this->t('Start typing a username to search for users. <strong>Please Note:</strong> When a user is selected from the options, <u>ONLY THE USER\'S ID WILL POPULATE</u> into this field.'),
            '#autocomplete_route_name' => 'stacked_permissions.search_user_autocomplete',
            '#autocomplete_route_parameters' => ['permission_group' => $this->entity->id()],
        ];

        // Add the "Add Member" button
        $form['add_member'] = [
            '#type' => 'submit',
            '#value' => $this->t('Add Member'),
            '#submit' => ['::submitAddMember'],
            '#ajax' => [
                'callback' => '::addMemberCallback',
                'wrapper' => 'members-table',
            ],
        ];

        // Insert a horizontal line followed by a header
        $form['manage_group_membership_separator'] = [
            '#markup' => '<h2>' . $this->t('Manage Group Membership:') . '</h2>',
        ];

        $form['search_member_list'] = [
            '#type' => 'textfield',
            '#title' => $this->t('Search Members in Group'),
            '#description' => $this->t('Enter all or part of a username to search current group members.'),
            '#size' => 30,
            '#attributes' => ['id' => 'search-member-list-input'],
        ];

        $form['search_button'] = [
            '#type' => 'submit',
            '#value' => $this->t('Search Members'),
            '#submit' => ['::searchMembers'],
            '#ajax' => [
                'callback' => '::ajaxRebuildTable',
                'wrapper' => 'members-table',
            ],
        ];

        // Add a container for messages.
        $form['custom_messages_remove'] = [
            '#type' => 'container',
            '#attributes' => ['id' => 'custom-messages-remove'],
        ];

        $form['group_membership_table_separator'] = [
            '#markup' => '<p><strong>' . $this->t('Current Members:') . '</strong></p>',
        ];

        $form['members'] = [
            '#type' => 'table',
            '#header' => [
                $this->t('Username'),
                $this->t('Operations'),
            ],
            '#rows' => $this->buildMembershipRows($form_state->getValue('search_member_list', '')),
            '#empty' => $this->t('No members found.'),
            '#pager' => ['#type' => 'tableselect', '#pager_id' => 'user-members-pagination', '#pages' => 50],
            '#attributes' => ['id' => 'members-table'],
        ];

        // Add a pager
        $form['membership-pager'] = [
            '#type' => 'pager',
        ];

        // Insert a horizontal line followed by a header
        $form['group_permissions_separator'] = [
            '#markup' => '<hr><h2>' . $this->t('Group Permissions:') . '</h2>',
        ];

        $form['permissions_list'] = $this->buildPermissionsList();

        return $form;
    }

    public function searchMembers(array &$form, FormStateInterface $form_state) {
        // Retrieve the search input from the form state.
        $search_term = $form_state->getValue('search_member_list', '');

//        // Log the search term for debugging purposes.
//        \Drupal::logger('stacked_permissions')->info('Searching members with term: ' . $search_term);

        // Find users matching the search term within the group.
        if (!empty($search_term)) {
            $user_ids = \Drupal::entityQuery('user')
                ->condition('name', '%' . $search_term . '%', 'LIKE')
                ->accessCheck(TRUE)
                ->execute();

            if (!empty($user_ids)) {
                $membership_storage = \Drupal::entityTypeManager()->getStorage('permission_group_membership');
                $memberships = $membership_storage->loadByProperties(['group_id' => $this->entity->id()]);

                $filtered_memberships = array_filter($memberships, function ($membership) use ($user_ids) {
                    return in_array($membership->get('user_id')->target_id, $user_ids);
                });

                // Update the form's members table with the filtered results.
                $form['members']['#rows'] = $this->buildMembershipRowsFromMemberships($filtered_memberships);
            } else {
                // If no user IDs match, clear the table.
                $form['members']['#rows'] = [];
            }
        }
    }

// Helper function to build rows from filtered memberships.
    private function buildMembershipRowsFromMemberships($memberships) {
        $rows = [];
        foreach ($memberships as $membership) {
            $user = $membership->get('user_id')->entity;
            if ($user) {
                $rows[] = [
                    'data' => [
                        $user->getDisplayName(),
                        $this->buildRemoveButton($user),
                    ],
                    'class' => ['member-row-' . $user->id()],
                ];
            }
        }
        return $rows;
    }

    public function ajaxRebuildTable(array &$form, FormStateInterface $form_state) {
        // This AJAX callback returns the updated members table with filtered results.
        return $form['members'];
    }

    public function submitAddMember(array &$form, FormStateInterface $form_state) {
        $response = new AjaxResponse();
        $user_id = $form_state->getValue('search_member');
//        \Drupal::logger('stacked_permissions')->info('Adding member with user ID: ' . $user_id);

        if (!empty($user_id)) {
            $user = User::load($user_id);

            if ($user) {
                $membership_storage = \Drupal::entityTypeManager()->getStorage('permission_group_membership');
                $existing_memberships = $membership_storage->loadByProperties([
                    'group_id' => $this->entity->id(),
                    'user_id' => $user_id,
                ]);

                if (empty($existing_memberships)) {
                    $membership = $membership_storage->create([
                        'user_id' => $user_id,
                        'group_id' => $this->entity->id(),
                    ]);
                    $membership->save();

                    \Drupal::messenger()->addStatus($this->t('User added to Permission Group successfully.'));
                } else {
                    \Drupal::messenger()->addError($this->t('This user is already a member of this group.'));
                }
            } else {
                \Drupal::messenger()->addError($this->t('Could not find the selected user to add to the Permission Group.'));
            }
        } else {
            \Drupal::messenger()->addError($this->t('Please find & select a user to add before selecting "Add Member".'));
        }

        return $response;
    }

    /**
     * Builds rows for the membership table.
     */
    private function buildMembershipRows($search = '') {
        $rows = [];
        $query = \Drupal::entityQuery('permission_group_membership')
            ->condition('group_id', $this->entity->id())
            ->accessCheck(TRUE); // Ensure access checking is applied

        // Add a search condition if a search term is provided
        if (!empty($search)) {
            $user_ids = \Drupal::entityQuery('user')
                ->condition('name', '%' . $search . '%', 'LIKE')
                ->accessCheck(TRUE) // Ensure access checking is applied
                ->execute();
            if (!empty($user_ids)) {
                $query->condition('user_id', $user_ids, 'IN');
            } else {
                // Return early if no users match the search
                return $rows;
            }
        }

        // Implement pagination by limiting the query
        $query->pager(50, 'membership_pager');
        $membership_ids = $query->execute();
        $memberships = \Drupal::entityTypeManager()->getStorage('permission_group_membership')->loadMultiple($membership_ids);

        foreach ($memberships as $membership) {
            $user = $membership->get('user_id')->entity;
            if ($user) {
                $rows[] = [
                    'data' => [
                        $user->getDisplayName(),
                        $this->buildRemoveButton($user),
                    ],
                    'class' => ['member-row-' . $user->id()],
                ];
            }
        }

        return $rows;
    }

    /**
     * Get the group members.
     */
    private function getGroupMembers() {
        $group = $this->entity;
        $membership_storage = \Drupal::entityTypeManager()->getStorage('permission_group_membership');

        // Load memberships associated with this group.
        $memberships = $membership_storage->loadByProperties(['group_id' => $group->id()]);
        $members = [];

        // Extract user entities from memberships.
        foreach ($memberships as $membership) {
            $user = $membership->get('user_id')->entity;
            if ($user) {
                $members[] = $user;
            }
        }

        return $members;
    }

    /**
     * Builds the remove button for each member.
     */
    private function buildRemoveButton($member) {
        $url = Url::fromRoute('stacked_permissions.remove_member', [
            'permission_group' => $this->entity->id(),
            'user' => $member->id(),
        ], ['attributes' => ['class' => ['use-ajax'], 'data-dialog-type' => 'ajax']]);

        return Link::fromTextAndUrl($this->t('Remove'), $url)->toString();
    }

    /**
     * Callback to handle adding a member.
     */
    public function addMemberCallback(array &$form, FormStateInterface $form_state) {
        $response = new AjaxResponse();

        // Render messages and add them to the custom message container.
        $messages = ['#type' => 'status_messages'];
        $status_messages = \Drupal::service('renderer')->renderRoot($messages);
        $response->addCommand(new HtmlCommand('#custom-messages-add', $status_messages));

        // Get the current input and rebuild the members table.
        $search = $form_state->getValue('search_member_list', '');
        $form['members']['#rows'] = $this->buildMembershipRows($search);

        // Return the updated members table.
        $members_table_rendered = \Drupal::service('renderer')->renderRoot($form['members']);
        $response->addCommand(new ReplaceCommand('#members-table', $members_table_rendered));

        // Clear the search_member field after adding a member.
        $response->addCommand(new \Drupal\Core\Ajax\InvokeCommand('#edit-search-member', 'val', ['']));

        return $response;
    }

    public function buildPermissionsList() {
        $config = \Drupal::config('stacked_permissions.allowed_permissions');
        $allowed_permissions = $config->get('permissions') ?: [];

        // Check if there are no allowed permissions configured.
        if (empty($allowed_permissions)) {
            return [
                '#markup' => $this->t('No permissions have been configured as approved to display on permission groups. Please contact your administrator so that approved permissions can be updated.'),
                '#prefix' => '<div class="no-permissions-message">',
                '#suffix' => '</div>',
            ];
        }

        // Retrieve permissions from the current PermissionGroup entity.
        $stored_permissions = unserialize($this->entity->get('permissions')->value) ?: [];

        $permission_handler = \Drupal::service('user.permissions');
        $permissions = $permission_handler->getPermissions();

        $grouped_permissions = [];
        $form = [];
        foreach ($permissions as $permission_id => $permission) {
            $module = $permission['provider'];
            try {
                $module_name = \Drupal::service('extension.list.module')->getName($module);
            } catch (\Exception $e) {
                $module_name = $module; // Fallback to the machine name
            }

            // Use -_- to replace spaces.
            $safe_permission_id = str_replace(' ', '-_-', $permission_id);
            $grouped_permissions[$module_name . '_|_' . $module][$safe_permission_id] = $permission;
        }

        foreach ($grouped_permissions as $module_name_concat => $permissions) {
            $module_name = explode('_|_', $module_name_concat)[0];
            $module = explode('_|_', $module_name_concat)[1];

            // Initialize children array to track if there are any allowed permissions.
            $children = [];
            foreach ($permissions as $permission_id => $permission) {
                $permission_key = $module . '_|_' . $permission_id;
                if (array_key_exists($permission_key, $allowed_permissions)) {
                    // Set default value based on stored permissions or default to 0 if not set
                    $default_value = isset($stored_permissions[$permission_key]) ? (bool)$stored_permissions[$permission_key] : FALSE;
                    $children[$permission_key] = [
                        '#type' => 'checkbox',
                        '#title' => $permission['title'],
                        '#default_value' => $default_value,
                    ];
                }
            }

            // Only add sections to the form if there are children.
            if (!empty($children)) {
                $form[$module_name] = [
                    '#type' => 'details',
                    '#title' => $this->t($module_name),
                    '#open' => FALSE,
                ];
                // Add children to the section.
                $form[$module_name] += $children;
            }
        }

        return $form;
    }

//    public function buildPermissionsList() {
//        $config = \Drupal::config('stacked_permissions.allowed_permissions');
//        $allowed_permissions = $config->get('permissions') ?: [];
//
//        // Check if there are no allowed permissions configured.
//        if (empty($allowed_permissions)) {
//            return [
//                '#markup' => $this->t('No permissions have been configured as approved to display on permission groups. Please contact your administrator so that approved permissions can be updated.'),
//                '#prefix' => '<div class="no-permissions-message">',
//                '#suffix' => '</div>',
//            ];
//        }
//
//        $permission_handler = \Drupal::service('user.permissions');
//        $permissions = $permission_handler->getPermissions();
//
//        $grouped_permissions = [];
//        $form = [];
//        foreach ($permissions as $permission_id => $permission) {
//            $module = $permission['provider'];
//            $module_name = \Drupal::service('extension.list.module')->getName($module);
//            // Use -_- to replace spaces.
//            $safe_permission_id = str_replace(' ', '-_-', $permission_id);
//            $grouped_permissions[$module_name.'_|_'.$module][$safe_permission_id] = $permission;
//        }
//
//        foreach ($grouped_permissions as $module_name_concat => $permissions) {
//            $module_name = explode('_|_', $module_name_concat)[0];
//            $module = explode('_|_', $module_name_concat)[1];
//
//            // Initialize children array to track if there are any allowed permissions.
//            $children = [];
//            foreach ($permissions as $permission_id => $permission) {
//                if (array_key_exists($module . '_|_' . $permission_id, $allowed_permissions)) {
//                    $children[$module . '_|_' . $permission_id] = [
//                        '#type' => 'checkbox',
//                        '#title' => $permission['title'],
//                        '#default_value' => false,
//                    ];
//                }
//            }
//
//            // Only add sections to the form if there are children.
//            if (!empty($children)) {
//                $form[$module_name] = [
//                    '#type' => 'details',
//                    '#title' => $this->t($module_name),
//                    '#open' => FALSE,
//                ];
//                // Add children to the section.
//                $form[$module_name] += $children;
//            }
//        }
//
//        return $form;
//    }

    /**
     * {@inheritdoc}
     */
    public function submitForm(array &$form, FormStateInterface $form_state) {
        // Call the parent submit handler first to process the entity form.
        parent::submitForm($form, $form_state);

        // Get all values from the form state.
        $values = $form_state->getValues();

        // Filter out fields not needed for storage.
        $fields_to_exclude = ['search_member', 'search_member_list', 'members', 'custom_messages_add', 'custom_messages_remove', 'add_member', 'search_button', 'membership-pager'];

        $filtered_values = array_filter($values, function ($key) use ($fields_to_exclude) {
            return !in_array($key, $fields_to_exclude, true);
        }, ARRAY_FILTER_USE_KEY);

        // Fetch all available permissions.
        $permission_handler = \Drupal::service('user.permissions');
        $all_permissions = $permission_handler->getPermissions();

        $selected_permissions = [];
        foreach ($filtered_values as $key => $value) {
            // Convert key back to original format and check if it's a permission
            $module = explode('_|_', $key)[0];
            $permission_id = str_replace('-_-', ' ', explode('_|_', $key)[1]);
            try {
                $module_name = \Drupal::service('extension.list.module')->getName($module);
            } catch (\Exception $e) {
                $module_name = '';
            }

            // Check if the permission exists and is checked.
            if (!empty($module_name) && array_key_exists($permission_id, $all_permissions)) {
                $selected_permissions[$module.'_|_'.str_replace(' ', '-_-', $permission_id)] = $value;
            }
        }

//        \Drupal::logger('stacked_permissions')->info('Selected permissions: @permissions', [
//            '@permissions' => json_encode($selected_permissions),
//        ]);

        // Serialize the permissions and save them in the entity.
        $this->entity->set('permissions', serialize($selected_permissions));
        $this->entity->save();

        // Invalidate this group's cache
        \Drupal::service('cache_tags.invalidator')->invalidateTags($this->entity->getCacheTags());
    }
}
